/* **********************************************************************************************************************
 * MetaData models for Form Fields
 * -------------------------------
 * Keep in one file for now, but maybe split if this grows too large
 ********************************************************************************************************************* */

import { TemplateRef } from '@angular/core'; 
import { ValidatorFn, AsyncValidatorFn } from '@angular/forms';
import { ValueTransformer } from './../interfaces';

interface SimpleFieldMetaData {
	name: string; 								// The FormControl name
	origin?: string;							// Location in API-returned model - defaults to name
	type?: string; 								// The component type e.g. BasicInput, Checkbutton, Timepicker, etc
	label?: string;								// The field label - defaults to unCamelCased name if not supplied
	value?: any;								// The field value - defaults to empty string if not supplied
	default?: any;								// Default value
	placeholder?: string;						// Optional placeholder text
	class?: string | Array<string>;				// CSS classes to apply
	id?: string;								// CSS id to apply
	isDisabled?: boolean;						// Whether field is initially disabled
	validators?: Array<ValidatorFn>;			// Array of validator functions - following Angular FormControl API
	asyncValidators?: Array<AsyncValidatorFn>;	// Array of async validator functions - following Angular FormControl API
	valFailureMsgs?: StringMap;					// Validation failure messages - display appropriate message if validation fails
}

interface Option {
	label: string;
	value: string | number | boolean;
}

interface OptionsFieldMetaData extends SimpleFieldMetaData {
	options;									// Array of Options - for select, radio-button-group and other 'multiple-choice' types
	horizontal?;								// Whether to arrang radio buttons or checkboxes horizontally (default false)
}
interface DropdownModifiedInputFieldMetaData extends SimpleFieldMetaData {
	modifiers: string[];
	transform: ValueTransformer;
}

interface TimePickerFieldMetaData extends SimpleFieldMetaData {
	// TODO: Tighhten up om types
	value: Date;
	format: string;
	steps: StringMap;
}

// ---------------------------------------------------------------------------------------------------------------------
// Form Field MetaData Models
// ---------------------------------------------------------------------------------------------------------------------
// Base Implementations

abstract class SimpleField {
	type = 'text';
	name = 'missingName';
	origin?: string;
	label?: string;
	value;
	default = '';
	placeholder = '';
	class?: string | Array<string>;
	id?: string;
	isDisabled?: boolean;
	validators: Array<ValidatorFn> = [];
	asyncValidators: Array<AsyncValidatorFn> = [];
	valFailureMsgs: StringMap = {};

	constructor(meta: SimpleFieldMetaData) {
		Object.assign(this, meta);
		if (!this.origin) {
			// If origin is not supplied it's the same as the name
			this.origin = this.name;
		}
		if (!this.label) {
			// If label is not supplied set it to the unCamelCased'n'Spaced name
			// e.g. supervisorCardNumber --> Supervisor Card Number
			this.label = this.name.replace(/([A-Z])/g, ' $1').replace(/^./, s => s.toUpperCase());
		}
	}
}

class Option {
	constructor(opt: string | Option) {
		if (typeof opt === 'object') {
			this.label = opt.label;
			this.value = opt.value;
		} else if (typeof opt === 'string') {
			this.label = opt;
			this.value = opt;
		}
	}
}

abstract class OptionsField extends SimpleField {
	options: Option[] = [];
	constructor(meta: OptionsFieldMetaData) {
		super(meta);
		if (Array.isArray(meta.options)) {
			this.options = meta.options.reduce((acc, opt) => { acc.push(new Option(opt)); return acc; }, []);
		} else {
			this.options = [
				new Option({ label: 'Yes', value: true }),
				new Option({ label: 'No', value: false })
			];
		}
	}
}

// ---------------------------------------------------------------------------------------------------------------------
// Concrete Implmemntations (suuplied for convenience)

class TextField extends SimpleField {
	type = 'Text';
}

class TextareaField extends SimpleField {
	type = 'Textarea';
}

class PasswordField extends SimpleField {
	type = 'Password';
}

class SelectField extends OptionsField {
	type = 'Select';
}

class RadioField extends OptionsField {
	type = 'Radio';
}

// ---------------------------------------------------------------------------------------------------------------------
// Custom Form Component Models

class CheckbuttonField extends SimpleField {
	type = 'Checkbutton';
	default: any = false;
}

class DropdownModifiedInputField extends SimpleField {
	type = 'DropdownModifiedInput';
	modifiers: string[];
	transform: ValueTransformer;
	constructor(meta: DropdownModifiedInputFieldMetaData) {
		super(meta);
	}
}

// ---------------------------------------------------------------------------------------------------------------------
// Custom FormGroup Component Models - render a series of FormControls

class CheckbuttonGroup {
	type = 'CheckbuttonGroup';
	name: string;
	label?: string;
	groupName: string;
	firstEnablesRest?;
	meta: CheckbuttonField[];
	constructor(groupmeta: any) {
		Object.assign(this, groupmeta);
		if (!this.label) {
			// If label is not supplied set it to the unCamelCased'n'Spaced name
			// e.g. supervisorCardNumber --> Supervisor Card Number
			this.label = this.name.replace(/([A-Z])/g, ' $1').replace(/^./, s => s.toUpperCase());
		}
		this.meta = [];
		while (groupmeta.meta.length) {
			this.meta.push( new CheckbuttonField(groupmeta.meta.shift()) );
		}
	}
}

// ---------------------------------------------------------------------------------------------------------------------
// Kendo Form Component Models

class TimepickerField<TimePickerFieldMetaData> extends SimpleField {
	type = 'Timepicker';
	value: Date = new Date();
	format = 'hh:mm a';
	steps = { hour: 1, minute: 15 };
}

class DatepickerField extends SimpleField {
	type = 'Datepicker';
	value: Date = new Date();
}

// ---------------------------------------------------------------------------------------------------------------------
// Container

class Container {
	type = 'Dynaform';
	label = '';
	template?: TemplateRef<any>;
	meta: StringMap; // TODO: Tighten up on type with union type
	constructor(children: StringMap) {
		this.meta = children;
	}
}

// ---------------------------------------------------------------------------------------------------------------------
// ---------------------------------------------------------------------------------------------------------------------
// ---------------------------------------------------------------------------------------------------------------------
// Exports

export {
	SimpleField,
	TextField, TextareaField, PasswordField, SelectField, RadioField,
	CheckbuttonField, DropdownModifiedInputField,
	CheckbuttonGroup,
	TimepickerField, DatepickerField,
	Container
};

/* *********************************************************************************************************************
 * Usage
 * Just ideas here for reference at the moment
 ******************************************************************************************************************** */

/*

 (1) Approach 1: Loop through model and 'lazily generate' the metadata. Something like...

 overrides = {
	 field3: {
		 type: Chechkbutton,
		 label: 'Friendly Name'
	 }
 }
modeledMeta = {};
for (field in model) {
	if (field in overrides) {
		Object.assign(modeledMeta, FieldFactory(model, overrides[field])); // FieldFactory returns a new field model of the specified type
	} else {
		Object.assign(modeledMeta, new SimpleField(field)); // Defaults to basic text field
	}
 }


 (2) Apprach 2: Meta-data fully specified (not lazily-generated), so loop through metadata genetating field models.
     metaspec may be hard-coded (for now) or loaded from server and cached, as part of app boostraping

 metaspac = {
	// Meta-data spec
 }
 modeledMeta = {}
 for (field in metaspec) {
	Object.assign(modeledMeta, FieldFactory(model, metaspec[field]));
 }



 ...
 Then pass modeledMeta to dynamic form generator / layout, etc

 */
