import { FormBuilder, FormGroup, FormControl } from '@angular/forms';
import { reduce } from 'lodash/fp';
import * as fmdModels from './../models';

/*
 * FORM METADATA UTILITIES
 *
 * Exports
 * -------
 * autoMeta(model) - generate basic metadata from a raw or mapped model, recursively if necessary
 * combineExtraMeta(metadata, extraMeta) - combine extra metadata into metatdata, lazyly and recursively
 * combineModelWithMeta(model, extraMeta) - automatically generated metadata for model then combines extra metadata
 * buildFieldSpecificMeta(metadata) - use field metadata models to expand metadata
 * buildFormGroupFunctionFactory(fb) - return a function to buildFormGroups using the supplied FormBuilder singleton
 *
 * Variable names
 * --------------
 * metaF   = metadata for Field
 * metaG   = metadata for Group (possibly nested)
 * metaFoG = metadata for Field Or Group
 *
 */

// ---------------------------------------------------------------------------------------------------------------------
// Raw Model (or Mapped Raw Model) to Automatic Metadata
// ---------------------------------------------------------------------------------------------------------------------

const isScalar = val => typeof val === 'boolean' || typeof val === 'number' || typeof val === 'string';

const keyValPairToMeta = (val, key) => ({ name: key, [isScalar(val) ? 'value' : 'meta']: val });
const keyValPairToMetaRecursive = ( [key, val] ) => {
	const innerVal = isScalar(val) ? val : autoMeta(val);
	const metaVal = keyValPairToMeta(innerVal, key);
	return [key, metaVal];
};
const autoMeta = model => Object.entries(model)
	.map(keyValPairToMetaRecursive)
	.reduce((res, [key, val]) => addProp(res, key, val), {});


// ---------------------------------------------------------------------------------------------------------------------
// Combine automatically generated metadata with overrides
// ---------------------------------------------------------------------------------------------------------------------

const combineMetaForField = (metaF, extraMetaF) => Object.assign(metaF, extraMetaF);
const combineExtraMeta = (metaG, extraMeta) => {
	const combinedMeta = {};
	Object.entries(extraMeta).forEach(([key, val]) => {
		if (typeof metaG[key] === 'object') {
			combinedMeta[key] = (<any>val).meta ?
				combineMetaForField(metaG[key], { meta: combineExtraMeta(metaG[key].meta, (<any>val).meta) }) :
				combineMetaForField(metaG[key], val);
		}
	});
	return { ...metaG, ...combinedMeta };
};

// Combine model with overrides (after automatically generating metadata from the model)
const combineModelWithMeta = (model, extraMeta) => combineExtraMeta(autoMeta(model), extraMeta);


// ---------------------------------------------------------------------------------------------------------------------
// Build Form-Field-Type-Specific Metadata (using the field models in dynaform/models)
// ---------------------------------------------------------------------------------------------------------------------

const buildFieldClassName = (t = 'text') => {
	const start = t[0].toUpperCase() + t.slice(1);
	if (start === 'Container' || t.slice(-5) === 'Group') {
		return start;
	}
	return start + 'Field';
};

const buildModeledField = metaFoG => {
	const type = isContainer(metaFoG) ? 'container' : metaFoG.type;
	const className = buildFieldClassName(type);
	if (!fmdModels[className]) {
		throw new Error(`No metadata model "${className}" for type "${type}"`);
	}
	return new fmdModels[className](metaFoG);
};

// Build Form Group Member
const buildModeledFieldGroupMember = metaFoG => {
	const modeledGroupMember = buildModeledField(metaFoG);
	if (isContainer(metaFoG)) {
		modeledGroupMember.meta = _buildModeledFieldGroup(metaFoG.meta);
	}
	return modeledGroupMember;
};

// Build Form Group
const buildModeledFieldGroupReducerIteree = (res, metaFoG) => Object.assign(res, { [metaFoG.name]: buildModeledFieldGroupMember(metaFoG) });
const _buildModeledFieldGroup = metaG => reduce(buildModeledFieldGroupReducerIteree, {}, metaG);
const buildFieldSpecificMeta = metaG => _buildModeledFieldGroup(metaG);


// ---------------------------------------------------------------------------------------------------------------------
// Build FormGroups
// ---------------------------------------------------------------------------------------------------------------------

const buildFormGroupFunctionFactory = (fb: FormBuilder): (meta) => FormGroup => {
	// Establishes a closre over the supplied FormBuilder and returns a function that builds FormGroups from metadata
	// ( it's done this way so we can use the FormBuilder singleton without reinitialising )

	// Build Form Control
	// TODO: Flesh out function to build validators
	const buildControlState = metaF => ({ value: metaF.value || metaF.default, disabled: metaF.isDisabled });
	const buildValidators = metaF => ({
		validators: null,
		asyncValidators: null,
		updateOn: 'blur'
	});
	const buildFormControl = metaF => new FormControl(buildControlState(metaF) /*, buildValidators(metaF) */);

	// Build Form Group Member
	const buildFormGroupMember = metaFoG => isGroup(metaFoG) ?
		_buildFormGroup(metaFoG.meta) :
		buildFormControl(metaFoG);

	const buildFormGroupReducerIteree = (res, metaFoG) => Object.assign(res, { [metaFoG.name]: buildFormGroupMember(metaFoG) });
	const _buildFormGroup = _metaG => fb.group(reduce(buildFormGroupReducerIteree, {}, _metaG));

	const buildFormGroup = metaG => {
		const metaWithNames = addMissingNames(metaG);
		return _buildFormGroup(metaWithNames);
	};
	return buildFormGroup;
};


// ---------------------------------------------------------------------------------------------------------------------
// Helper Funstions
// ---------------------------------------------------------------------------------------------------------------------

// Add Property to object
const addProp = (obj, key, val) => { obj[key] = val; return obj; };

// Is Group
// Helper function to distinguish group from field
const isGroup = metaFoG => metaFoG.meta;

// Is Container
// Helper function to distinguish container group (a group of child fields)
const isContainer = metaFoG => isGroup(metaFoG) && (metaFoG.type === 'container' || typeof metaFoG.type === 'undefined');

// Add Missing Names
// Helper function to add any missing 'name' properties to Fields and Groups using property's key, recursively
const addNameIfMissing = (metaFoG, key) => metaFoG.name ? metaFoG : addProp(metaFoG, 'name', key);
const addNameToSelfAndChildren = ( [key, metaFoG] ) => {
	metaFoG = addNameIfMissing(metaFoG, key);
	if (isGroup(metaFoG)) {
		metaFoG.meta = addMissingNames(metaFoG.meta); // Recursion
	}
	return [key, metaFoG];
};
const addMissingNames = metaG => Object.entries(metaG)
	.map(addNameToSelfAndChildren)
	.reduce((res, [key, val]) => addProp(res, key, val), {});


// ---------------------------------------------------------------------------------------------------------------------
// Exports
// ---------------------------------------------------------------------------------------------------------------------

export { autoMeta, combineModelWithMeta, combineExtraMeta, buildFieldSpecificMeta, buildFormGroupFunctionFactory };
